//
//  Microware logics: hierarchical State Machine
//  Guit GUI Toolkit
//  Copyright © 2019/2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <guit.hpp>
#include <gstatechart.hpp>
#include <gshape.hpp>
#include "examples.hpp"
using namespace std;
using namespace guit;


class Microwave : public GWindow {
  int default_duration = 10;

  // for each GInt: initial value, min value, max value
  GInt duration{default_duration,0,200}, power{1,1,10}, mode{0,0,2}, weight{0,0,100};
  GStrings modes{"Waves", "Grill", "W+G"};

  GStateChart sc;          // statechart
  GSuperState init{sc};    // hierarchical state (parent is statechart)
  // other states (parent is init hierarchical state)
  GState idle{init}, cooking{init}, set_hours{init}, set_minutes{init},
  set_power{init}, set_mode{init}, set_defrost{init}, set_duration{init};

  // Clock and cooking timers: Fired every second indefinitely
  GTimer clocktimer{1000}, cooktimer{1000};
  GInt seconds, minutes, hours;

public:
  Microwave() : GWindow("#microwave Microwave Oven")
  {
    // Count the number of seconds, minutes and hours. For instance, seconds is
    // increased every time the timer is fired (= every second) until it reaches 60.
    // It is then reset to 0 and so on.
    clocktimer.addBeat(seconds, 1, 60);
    clocktimer.addBeat(minutes, 60, 60);
    clocktimer.addBeat(hours, 60*60, 24);

    // Displays time or current mode depending on the current state:
    // - when idle entered:       "~hours : ~minutes : ~seconds"
    // - when set_hours entered:  "Hour: ~hour"
    // - etc.
    // ~hours, etc. will be the (future) value of hours, etc.
    auto& display = Label("#display {format: 2 2 0}")  // precision width fill
    << +idle           / (~hours + " : " + ~minutes + " : " + ~seconds)
    << +set_hours      / ("Hour: " + ~hours)
    << +set_minutes    / ("Minutes: " + ~minutes)
    << +set_power      / ("Power: " + ~power)
    << +set_mode       / ("Mode: " + (~modes)[~mode])
    << +set_defrost    / ("Weight: " + ~weight)
    << (+set_duration) / ("Time: " + ~duration)
    << +cooking        / ("Cooking: " + (~duration - ~cooktimer)+"s");


    // Contains the + and - dial buttons. on::action is the default action
    // (clicking the button), [] is a guard so that only the action corresponding
    // to the valid guardwill be performed, e.g.:
    // - when button + clicked, if set_hours entered: executes hours++
    // - when button - clicked, if set_hours entered: executes hours--
    auto& controls = ButtonBox("#controls")
    << (Button("Clock")
        << (idle >> set_hours)
        << (set_hours >> set_minutes)
        << (set_minutes >> idle)
        )
    << (Button("Power")
        << (idle >> set_power)
        << (set_power >> set_duration)
        )
    << (Button("Mode")
        << (idle >> set_mode)
        << (set_mode >> set_duration)
        )
    << (Button("Defrost")
        << (idle >> set_defrost)
        )
    << (Button(".dialbtn - ")
        << Guard(+set_hours)   / [&]{hours--;}
        << Guard(+set_minutes) / [&]{minutes--;}
        << Guard(+set_power)   / [&]{power--;}
        << Guard(+set_mode)    / [&]{mode--;}
        << Guard(+set_defrost) / [&]{weight--;}
        << Guard(+set_duration || +cooking) / [&]{duration--;}
        )
    << (Button(".dialbtn + ")
        << Guard(+set_hours)   / [&]{hours++;}
        << Guard(+set_minutes) / [&]{minutes++;}
        << Guard(+set_power)   / [&]{power++;}
        << Guard(+set_mode)    / [&]{mode++;}
        << Guard(+set_defrost) / [&]{weight++;}
        << Guard(+set_duration || +cooking) / [&]{duration++;}
        )
    << (Button("Stop")
        << (init >> init)
        )
    << (Button("Start")
        << (idle >> cooking)    / [&]{duration = default_duration;}
        << (cooking >> cooking) / [&]{duration += default_duration;}
        << (set_duration >> cooking)
        << (set_defrost >> cooking)
        )
    // when duration elapsed => leave cooking state and enter start state
    << (~cooktimer >= ~duration) / (cooking >> init)

    // cooking entered => start timer and start cooking
    << +cooking / [&]{ cooktimer.start(); cook(true); }

    // cooking exited  => stop timer and stop cooking
    << -cooking / [&]{ cooktimer.stop(); cook(false); };


    // animation: changes Background from a color to another in 2000ms; loops infinitely
    auto& cooking_anim = Anim(2000)
    << Background("pink") << Background("firebrick");

    
    *this
    << on::show / [&]{clocktimer.start();}
    << on::hide / [&]{clocktimer.stop();}
    << (HBox
        // the animation runs while the 'cooking' state is entered.
        << (Label("#oven")
            << +cooking / cooking_anim
            )
        << (VBox("#panel {flex: 0 0}")
            << display
            << controls
            )
        );

    init.setInitial(idle);
    sc.setInitial(init);
    sc.start();            // starts the state machine
  }

  void cook(bool state) {
    if (state) std::cout << "Cooking..." << std::endl;
    else std::cout << "Idle" << std::endl;
  }
};


GWindow& Examples::makeMicrowave() {
  return *new Microwave;
}
