//
//  Font base
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
#include <unordered_map>
#include <filesystem>
#include <gapp.hpp>
#include <gstring.hpp>
#include <gfile.hpp>
#include <gfont.hpp>
#include <gfontbase.hpp>
#include <ggraphics.hpp>
#include <gcontrol.hpp>
#include <gnative.hpp>
using namespace std;
namespace fs = std::filesystem;
namespace guit {

#define FONT_LIST "font-list"

static std::unordered_map<GString,GFontFace*>* fontfaces_;
static GFontFace* fallbackface_;

static GString fontFaceId(GString const& family, GString const& style) {
  return strlower(family) + ":" + strlower(style);
}

static GString fontId(GString const& family, GString const& style, uint8_t size) {
  return strlower(family) + ":" + strlower(style) + ":" + ::to_string(size);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GFontFace::GFontFace(GString const& family, GString const& style, GString const& path)
: family_(family), style_(style), path_(path) {}

void GFontBase::init() {
  fontfaces_ = new std::unordered_map<GString,GFontFace*>;

  auto guitdir = GApp::guitDir();
  if (guitdir.empty()) throw GFatalError("GFontBase", "Guit directory couln't be retrieved");

  // dont rebuild the fontlist if it already exists
  if (!fileExist(guitdir+FONT_LIST)) makeFontList(guitdir);

  readFontList(guitdir);
}


bool GFontBase::makeFontList(GString const& guitdir) {
  //auto fontdirlist = guitdir + "font-dirs.txt";
  auto fontlist = guitdir+FONT_LIST;

  GStrings fontdirs = GPrefs::fontDirs();
  /*
  std::ifstream fd(fontdirlist);
  if (!fd) fontdirs = GPrefs::fontDirs();       // TODO!!!
  else {
   //raedThem(fontdirs);
   fontdirs = GPrefs::fontDirs();
   }*/

  if (fontdirs.empty()) {
    std::clog
    << "Calling fc-list to create the font list file: " << fontlist << "\n"
    << "You can edit this file if needed\n"
    << "If you add new font files, just remove this file, it will be created again automatically\n"
    << std::endl;

    system("fc-list > " + fontlist);
    std::ifstream in(fontlist);
    GString line;
    if (in && getline(in, line)) return true;
    else {
      gerror("GFontBase::makeFontList",
             "fc-list failed creating the font list file " + fontlist);
      system("rm " + fontlist);
      return false;
    }
  }

  std::ofstream flist(fontlist);
  if (!flist) {
    gerror("GFontBase::makeFontList",
           "Couln't create the font list file " + fontlist);
    return false;
  }
  std::clog
  << "Creating the font list file: " + fontlist <<  "\n"
  << "You can edit this file if needed\n"
  << "If you add new font files, just remove this file, it will be created again automatically\n"
  << std::endl;

  int count = 0;
  for (auto& dirname : fontdirs) {
    try {
      for (auto& p : fs::directory_iterator(dirname)) {
        GString family, style;
        if (GNatContext::getFontInfo(p.path().string(), family, style)
            && !family.empty()  && !style.empty()
            ) {
          count++;
          flist << p.path().string() << ": " << family << ":style=" << style << '\n';
        }
      }
    }
    catch (fs::filesystem_error& e) {
      gerror("GFontBase::makeFontList", GString("File system error: ") + e.what());
    }
  }
  flist << std::endl;


  if (count > 0) return true;
  else {
    gerror("GFontBase::makeFontList","No valid font was found, text won't be displayed");
    return true;
  }
}


bool GFontBase::readFontList(GString const& guitdir) {
  auto fontlist = guitdir+FONT_LIST;

  std::ifstream in(fontlist);
  if (!in) {
    gerror("GFontBase::readFontList", "Couln't open the font list file " + fontlist);
    return false;
  }
  GString line;
  int count = 0;

  while (std::getline(in, line)) {
    if (strblank(line) || line[0] == '#') continue;  // comment
    if (!addFontFace(line)) continue;
    if (in.fail()) {
      gerror("GFontBase::readFontList","premature EOF in font list file " + fontlist);
      return count;
    }
    count++;
  }

  if (count > 0) return true;
  else {
    gerror("GFontBase::readFontList",
           "No valid font was found in: "+ fontlist + ", text won't be displayed");
    return true;
  }
}


bool GFontBase::addFontFace(GString const& fontspec) {
  auto stylepos = fontspec.find(":style=");
  if (stylepos == GString::npos) return false;

  auto fampos = fontspec.rfind(": ",stylepos);
  if (fampos == GString::npos) return false;

  // Note: beware that Windows paths can start with C: or similar
  GString
  path(fontspec.substr(0, fampos)),
  family(fontspec.substr(fampos+2, stylepos-fampos-2)),
  style(fontspec.substr(stylepos+7));

  // cerr << "'"<< path << "'  '" << family << "'  '" << style << endl;

  if (path.empty() || family.empty() || style.empty() || !fileExist(path)) {
    gerror("GFontBase::addFontFace", "font entry invalid or not found: "+fontspec);
    return false;
  }

  auto fontid = fontFaceId(family, style);
  auto it = fontfaces_->find(fontid);

  if (it != fontfaces_->end()) return false;  // already found
  else {
    auto* f = new GFontFace(family, style, path);
    (*fontfaces_)[fontid] = f;
    if (!fallbackface_) fallbackface_ = f;
    return true;
  }
}


GFontFace* GFontBase::findFontFace(GString const& family, GString const& style) {
  if (!fontfaces_) init();

  if (family == "sans-serif") {
    for (auto& it : GPrefs::sansFonts()) {
      if (auto* f = findFontFace(it, style)) return f;
    }
    return nullptr;
  }
  else if (family == "serif") {
    for (auto& it : GPrefs::serifFonts()) {
      if (auto* f = findFontFace(it, style)) return f;
    }
    return nullptr;
  }
  else if (family == "monospace") {
    for (auto& it : GPrefs::monoFonts()) {
      if (auto* f = findFontFace(it, style)) return f;
    }
    return nullptr;
  }
  else {
    auto it = fontfaces_->find(fontFaceId(family, style));
    // cout << "- searching: " << family << " / style: " << style
    // << "' found: " << int(it != fontfaces_->end()) <<  endl;
    return it == fontfaces_->end() ? nullptr : it->second;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GFontResource::~GFontResource() {
  cerr << "GFontResource::~GFontResource" << endl;                 // TODO !!!!
}

GFontResource::GFontResource(GFontFace* face, GString const& style,
                             uint8_t size, GString const& families)
: size_(size), face_(face), style_(style), families_(families) {}

void GFontResource::set(GFontResource*& to, GFontResource* from) {
  if (to) to->refcount_--;
  to = from;
  if (to) to->refcount_++;
}

GFontResource* GFontResource::obtainFont(GString const& style, uint8_t size,
                                         GString const& families) {
  // !! MUST BE HERE bec this fun may be called before start() !!
  static std::unordered_map<GString,GFontResource*> fontimpls_;

  GFontFace* face{};
  GString family;
  size_t pos = 0;

  while (strtoken(families, family, pos, ",") != -1) {
    auto it = fontimpls_.find(fontId(family, style, size));
    if (it != fontimpls_.end()) return it->second;

    if ((face = GFontBase::findFontFace(family, style))) break;
  }

  if (!face) {
    if (!fallbackface_) gerror("GFontResource", "No font available");
    else {
      gerror("GFontResource","Font '" + families + " style: " + style
             +"' not found, using default font");
      face = fallbackface_;
    }
  }

  return fontimpls_[fontId(family, style, size)] = new GFontResource(face, style,
                                                                     size, families);
}


GFontResource* GFontResource::emptyFont() {
  static auto* f = new GFontResource(nullptr,"",0,"<no-font>");
  return f;
}

bool GFontResource::realize() {
  if (natfont_) return true;  // already set
  if (cantcreate_) return false;   // can't do anything, don't display a message again

  if (GNatContext::loadFont(*this)) return true;  // inits natfont_
  cantcreate_ = true;
  gerror("GFont::realize","Can't load font: " + families() + " " + style());
  return false;
}

}
