//
//  Boxes with a radio behavior
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gevent.hpp>
#include <gchoicebox.hpp>
#include <gtextitem.hpp>
#include <gborder.hpp>
#include <gicon.hpp>
#include <gmenu.hpp>
#include <gwrapper.hpp>
#include <gcontrol.hpp>
#include <gstyle.hpp>
#include <grender.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(ChoiceBox)
GIMPLEMENT_GADGET(ListBox)
GIMPLEMENT_GADGET(RadioBox)

GChoiceBox::GChoiceBox(GadgetType& t, GString const& spec) : GBox(t, spec) {
  addProp(*new GChoice, true);
}

GChoiceBox::GChoiceBox(GString const& spec)
: GChoiceBox(Type(), spec) {}

GChoiceBox::~GChoiceBox() = default;  // to avoid header inclusions (bec. of index_)

GChoiceBox& GChoiceBox::clone(GClone const& c) {
  auto* g = new GChoiceBox;
  //if (c & GClone::ShareProps) box->choice_ = choice_; // see initClone
  g->initClone(*this, c);
  return *g;
}

void GChoiceBox::initClone(Gadget& from, GClone const& c) {
  GBox::initClone(from, c);
  if (c & GClone::ShareProps) {
    if (auto* f = from.toChoiceBox()) choice_ = f->choice_;
  }
}

Gadget& GChoiceBox::createDefaultItem(GString const& arg) {
  return *new GItem(arg);
}

void GChoiceBox::setChoiceImpl(GChoice& c, bool add_to_list) {
  modes_.hasChoice = 1;
  choice_ = c;
  addProp2(c, add_to_list);
  // updates box when changed
  c.startNotify(*this, [&c](auto& g){c.radioCB(*g.toBox());});
}

void GChoiceBox::addProp(GProp& prop, bool add_to_list) {
  switch (prop.role()) {

    case GPropType::Choice:
      if (GChoice* c = prop.toChoice()) {
        setChoiceImpl(*c, add_to_list);
        c->fireDefaultCallbacks();
        updateLayout();
      }
      break;

    case GPropType::Text:
      if (auto* m = prop.toMutableProp()) {
        addProp2(prop, add_to_list);
        // box updates textprop
        auto* t = prop.toText();
        // list updates text
        *this << [this,t]{
          // if no match dont destroy the string
          if (chosen()) *t = chosenText();
        };
        *t = this->chosenText();
        // text  updates box
        m->startNotify(*this, [this,t](auto&){this->choose(t->value());});
      }
      else setValue(prop.stringValue());
      update();
      return;
      
    case GPropType::Expr:
    case GPropType::ExprWrapper: {
      GExprWrapper* w{};
      if (auto* e = prop.toExpr()) w = new GExprWrapper(*e);
      else if (!(w = prop.toExprWrapper())) return;
      
      auto* t = findProp<GText>();
      if (!t || t->isMutable()) add(*(t = new GText));
      w->bindWith(*t);
      addProp2(*w, add_to_list);
      updateLayout();
    } return;

    default:
      GBox::addProp(prop, add_to_list);
      return;
  }
}

GChoice& GChoiceBox::choice() {return *choice_;}

int GChoiceBox::value() const {return choice_->value();}

GString GChoiceBox::stringValue() const {return choice_->stringValue();}

int GChoiceBox::chosenIndex() const {return choice_->value();}

GString GChoiceBox::chosenText() const {
  Gadget* it = chosen();
  if (it) return it->text();
  else return emptystring_;
}

Gadget* GChoiceBox::chosen() const {
  return choice_->chosenItem(itemBox());
}

void GChoiceBox::choose(int index) {
  choice_->set(index);
}

void GChoiceBox::choose(GString const& text) {
  choice_->choose(itemBox(), text);
}

void GChoiceBox::choose(Gadget* item) {
  if (choiceBehavior() == ChoiceBehavior::Select) {
    if (item && !item->isCheckable()) {
      //error("select", "Gadget given as argument is not selectable");
      item->setCheckable(true);
    }
  }
  choice_->choose(itemBox(), item);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GListBox::GListBox(GadgetType& t, GString const& spec)
: GChoiceBox(t, spec) {}

GListBox::GListBox(GString const& spec)
: GChoiceBox(Type(), spec) {}

GListBox& GListBox::clone(GClone const& c) {
  auto* g = new GListBox();
  // if (c & GClone::ShareProps) g->choice_ = choice_; done initClone()
  g->initClone(*this, c);
  return *g;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GRadioBox::GRadioBox(GadgetType& t, GString const& spec)
: GChoiceBox(t, spec) {}

GRadioBox::GRadioBox(GString const& spec)
: GChoiceBox(Type(), spec) {}

GRadioBox& GRadioBox::clone(GClone const& c) {
  auto* g = new GRadioBox();
  // if (c & GClone::ShareProps) g->choice_ = choice_; done initClone()
  g->initClone(*this, c);
  return *g;
}


}

