//
//  Dialog boxes.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gapp.hpp>
#include <gevent.hpp>
#include <gexpr.hpp>
#include <gdialog.hpp>
#include <gitem.hpp>
#include <gtextitem.hpp>
#include <gcontrol.hpp>
#include <gstyle.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

class GDialogButtons : public GButtonBox {
public:
  GDialogButtons(GString const& s = "") : GButtonBox(Type(), s) {}
  using GType = GadgetType_<GDialogButtons,GButtonBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};

GDialogButtons& DialogButtons(GString const& css_and_title);
GIMPLEMENT_GADGET(DialogButtons)
GIMPLEMENT_GADGET(Dialog)
GIMPLEMENT_GADGET_TYPE(InfoDialog)
GIMPLEMENT_GADGET_TYPE(QuitDialog)

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int GDialog::confirm(GString const& message, GString const& info, GString const& buttons) {
  static auto& alert = *new GInfoDialog("Alert");
  alert.setInternal();
  alert.ignoreSmartPointers();
  alert.message() << message;
  alert.info() << (info.empty() ? "" : info);
  if (buttons.empty()) alert.setButtons("Cancel,OK"); else alert.setButtons(buttons);
  return alert.open();
}

int GDialog::alert(GString const& message, GString const& info) {
  static auto& alert = *new GInfoDialog("Alert");
  alert.setInternal();
  alert.ignoreSmartPointers();
  alert.message() << message;
  alert.info() << (info.empty() ? "" : info);
  alert.setButtons("OK");
  return alert.open();
}

GDialog::GDialog(GadgetType& t, GString const& spec) :
GWindow(t, spec) {}

GDialog::GDialog(GString const& spec) :
GWindow(Type(), spec) {}

GDialog& GDialog::clone(GClone const& cm) {
  auto* g = new GDialog;
  g->initClone(*this, cm);
  return *g;
}

GDialog& GDialog::show(bool state) {
  if (state) setResult(0, false);
  GWindow::show(state);
  return *this;
}

int GDialog::open() {
  setModal(true);
  show(true);
  quitloop_ = false;
  GCore::subLoop(quitloop_);
  return result();
}

void GDialog::setResult(int result, bool hide) {
  result_ = result;
  quitloop_ = hide;
  if (result == 1) {
    GEvent e(on::action, GCore::elapsedTime());
    e.window_ = this;
    e.gadget_ = this;
    props_.fire(e);
  }
  if (hide) close();
}

GBox& GDialog::buttonBox() {
  if (!bbox_) {
    bbox_ = new GDialogButtons();
    add(*bbox_);
    *bbox_ << [this](GEvent& e){
      auto* btn = e.chosen() ? e.chosen() : nullptr;
      setResult(btn ? bbox_->childPos(btn) : 0, true);
    };
  }
  return *bbox_;
}

void GDialog::setButtons(GString const& buttons) {
  auto& bbox = buttonBox();  // creates if neeeded
  bbox.removeChildren();     // removes existing buttons
  GStrings v;
  strsplit(buttons, v, ",");
  for (auto& it : v) bbox.add(Button(it));
}

GItem* GDialog::findButton(int pos) {
  auto* g = bbox_ ? bbox_->childAt(pos) : nullptr;
  return g ? g->toItem() : nullptr;
}

Gadget& GDialog::cancelButton() {
  auto& bbox = buttonBox();
  if (bbox.childCount() == 0) bbox.add(Button("Cancel"));
  return *bbox.childAt(0);
}

Gadget& GDialog::okButton() {
  auto& bbox = buttonBox();
  if (bbox.childCount() == 0) bbox.add(Button("Cancel"));
  if (bbox.childCount() == 1) bbox.add(Button("OK"));
  return *bbox.childAt(1);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GInfoDialog::GInfoDialog(GadgetType& t, GString const& spec) :
GDialog(t, spec),
message_(Label("/Message")) {
  setModal(true);
  message_->setMultiLine();
  *this << *message_ << Filler() << Separator();
  setButtons("Cancel,OK");
}

GInfoDialog::GInfoDialog(GString const& spec) :
GInfoDialog(Type(), spec) {}

GInfoDialog& GInfoDialog::clone(GClone const& c) {
  auto* g = new GInfoDialog;
  g->initClone(*this, c);
  return *g;
}

GItem& GInfoDialog::message() {
  return *message_;
}

GItem& GInfoDialog::info() {
  if (!info_) {
    info_ = Label();
    info_->setMultiLine();
    add(*info_, 1);
  }
  return *info_;
}

GTextItem& GInfoDialog::input() {
  if (!input_) {
    input_ = TextField();
    add(*input_, 1);
  }
  return *input_;
}

GInfoDialog& GInfoDialog::openOnlyIf(GBool& test) {
  openonly_ = &test;
  return *this;
}

GInfoDialog& GInfoDialog::openOnlyIf(GExprBool const& test) {
  openonly_ = &test.prop;
  return *this;
}

GInfoDialog& GInfoDialog::openOnlyIf(std::function<bool()> test) {
  openonly_ = &GExprBool(test).prop;
  return *this;
}

GInfoDialog& GInfoDialog::show(bool state) {
  if (state && openonly_ && !openonly_->boolValue()) {
    setResult(1, false);
  }
  else {
    if (state) setResult(0, false);
    GWindow::show(state);
  }
  return *this;
}


GInfoDialog& InfoDialog(GString const& title, GString const& mesg) {
  auto& d = *new GInfoDialog(title);
  if (!mesg.empty()) d.message() << mesg;
  return d;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GQuitDialog::GQuitDialog(GadgetType& t, GString const& s)
: GInfoDialog(t, s) {}

GQuitDialog::GQuitDialog(GString const& s)
: GInfoDialog(Type(), s) {}

GQuitDialog& GQuitDialog::clone(GClone const& cm) {
  auto* g = new GQuitDialog;
  g->initClone(*this, cm);
  return *g;
}

GQuitDialog& QuitDialog(GString const& title, GString const& mesg) {
  auto& d = *new GQuitDialog(title.empty() ? "Quit" : title);
  d.message() << (mesg.empty() ? "Do you want to quit?" : mesg);
  // d.info() << "Changes may be lost!";
  d.setButtons("Cancel,Quit");

  //d << []{GCore::exit(0);};
  d << [&d]{
    Gadget* par = d.getPseudoParent();
    GWindow* winpar = par ? par->getWindow() : nullptr;
    //cerr << "quit ";
    //if (winpar) cerr << winpar << " " << winpar->text();
    //cerr << endl;

    if (!winpar) GCore::exit(0);
    else {
      winpar->wmodes_.quitcalled = 1;
      winpar->close();
    }
  };

  return d;
}

}
