//
//  File Dialog
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_FileDialog_hpp
#define Guit_FileDialog_hpp
#include <gdialog.hpp>
namespace guit {

/// Dialog for selecting a file.
/// By default, an GFileDialog is modal (see setModal()) and has a Cancel
/// and a OK button.
/// @see DialogGadget.
class GFileDialog : public GDialog {
protected:
  GFileDialog(GadgetType&, GString const& css_and_title);

public:
  GFileDialog(GString const& css_and_title = "");

  GFileDialog& clone(GClone const&) override;

  using GType = GadgetType_<GFileDialog,GDialog>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// returns the pathname of the selected file.
  /// @{
  GString pathname() const;
  GString stringValue() const override {return pathname();}
  /// @}

  /// returns the name of the selected file without the directory.
  GString basename() const;

  /// returns the selected directory.
  GString dirname() const;

  /// changes the directory.
  /// "~" is the home dir, ":" the resource dir, ".." goes to the parent dir.
  GFileDialog& setDir(GString const& path);

  /// Shows/hides the dialog.
  /// @see open() and close() in superclasses.
  GFileDialog& show(bool state = true) override;

  /// Display an alert before opening the dialog if a condition is met (useful for OpenDialog(s)).
  /// If test() is true, an alert is displayed, typically to warn the user that the data has not been saved.
  /// @{
  GFileDialog& alertIf(GBool& test);
  GFileDialog& alertIf(GExprBool const& test);
  GFileDialog& alertIf(std::function<bool()> test);
  /// @}

  /// Displays an alert if the selected file already exists (useful for SaveDialog(s)).
  /// If the selected file already exists, an alert is displayed to warn the user that this file
  /// will be overwritten if he proceeds.
  GFileDialog& alertFileExist(bool state);

  GFileDialog& alertFileDontExist(bool state);

  /// returns the prop that controls the selection.
  GChoice& choice() {return *choice_;}

  /// @internal
  /// @{
protected:
  gptr<GChoice> choice_;
  gptr<GBox> list_;
  gptr<GTextItem> file_, dir_;
  gprotect<GProp> alertif_;
  bool alert_file_exist_{}, alert_file_dont_exist_{};
  void selectEntry(GString fname);
  void dblClick(bool isdir);
  void setResult(int result, bool hide) override;
  void rescan();
  /// @}
};

/// Creates a FileDialog (see GFileDialog).
GFileDialog& FileDialog(GString const& css_and_title = "");


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


class GOpenDialog : public GFileDialog {
public:
  GOpenDialog(GString const& css_and_title = "");

  GOpenDialog& clone(GClone const&) override;

  using GType = GadgetType_<GOpenDialog,GFileDialog>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates an OpenDialog.
GOpenDialog& OpenDialog(GString const& title = "", GString const& directory = "");


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


class GSaveDialog : public GFileDialog {
public:
  GSaveDialog(GString const& css_and_title = "");

  GSaveDialog& clone(GClone const&) override;

  using GType = GadgetType_<GSaveDialog,GFileDialog>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a SaveDialog.
GSaveDialog& SaveDialog(GString const& title = "", GString const& directory = "");


}
#endif
