//
//  Item Gadgets: labels, buttons and menu/list items.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Item_hpp
#define Guit_Item_hpp
#include <gadget.hpp>
namespace guit {

/// Base class for labels, state and stateless buttons, menu items, etc.
/// A GItem auto-adapts to its parent, meaning that its look & feel depends on its parent.
/// For instance it will look and behave as a menu item in a GMenu, as a GButton in a GButtonBox,
/// as a GRadioItem in a GRadioBox, as a GListItem in a GListBox/GOptionBox/GComboBox.
///
/// Subclasses of GItem (e.g. GCheckItem, GRadioItem) implement a more specific look & feel.
/// Some other subclasses implement gadgets that can have a numerical value (e.g. GValueitem,
/// GChartItem, GSlider, GScrollBar) or can handle text editing (GTextItem, GTextField, GTextArea)
///
/// A GItem is very similar to a Gadget. The main difference is that it can display rich text
/// (including images) and an icon. Text is not editable (see GItextItem for that purpose).

/// @section Text
/// - The text is encoded in UTF8.
/// - It is not editable interactivelly. Use TextField or TextArea for editable text.
/// - It can be changed by adding a String with operator <<
/// - It can be retreived by calling the text() method
/// - Internally an Item uses a GText prop to display the text. Adding a GText
///   changes that prop.
/// @code
///    auto& b1 = Button("toto");
///    b1 << "DoIt";              // changes the text
///    std::cout << b1.text();    // displays "Do It"
///
///    auto& txt = Text("Do It");
///    auto& b2 = Button("toto");
///    b2 << txt;                 // changes the GText prop, thus the text
///    txt = "Whatever";          // changes the text of txt and b2
///    std::cout << b2.text();    // displays "Whatever"
/// @endcode
///
/// @section Checking the item
/// see Gadget::check(), Gadget::isCkecked(), etc.
///
/// @section Binding Props
/// Props can be bound together:
/// @code
///    auto& value = Float();
///    auto& lbl   = Label << value;
///    auto& tf    = TextField << value;
///    auto& sl    = Slider << value;
/// @endcode
/// The label and the textfield will then display a float number and the user
/// will be forced to enter a number in the textfield.
///
/// Props can be bound together, see GVarProp.
/// Active expressions allow binding props and performing calculations, see GExpr.
///
/// Opening a Menu or a Window<br>
/// A Menu, a Window, a Dialog and subclasses can be added to an item.
/// The menu is automatically opened when the item is pressed, and the
/// window/dialog box when the item is clicked.
/// Other gadgets cannot be added to an item.
///
/// @see Gadget, GExpr, GText
///
class GItem : public Gadget {
protected:
  GItem(GadgetType& t) : Gadget(t) {}
  GItem(GadgetType&, GString const& css_and_title);

public:
  ~GItem();
  explicit GItem(GString const& css_and_title = "");

  GItem& clone(GClone const&) override;
  GItem* toItem() override {return this;}

  /// Returns the text of the item.
  /// @note The text can be changed as follows:
  /// @code
  ///    auto& btn = Button("toto");
  ///    btn << "DoIt";   // changes the text
  /// @endcode
  GString text() const override;

  /// Changes the text.
  ///@{
  virtual GItem& operator=(GString const& str);
  virtual GItem& operator=(float);
  ///@}

  GItem& setMultiLine(bool value = true);

  /// Perform on::action callbacks (the defaut action).
  virtual void doAction();

  /// returns "true" if the item is selected, "false" otherwise.
  GString stringValue() const override;
  
  /// selects or deselects the item if Gadget::istCheckable() is true.
  void setValue(GString const& text) override;
  
  bool hasValue() const override {return isCheckable();}

  GProp* valueProp() override;

  using GType = GadgetType_<GItem,Gadget>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  using Gadget::add;
  void add(GWindow*) override;
  void add(GMenu*) override;
  void addProp(GProp&, bool add_to_list) override;
  void pack(GRender* parent_render) override;
  class GTextCache* tcache_{};
  /// @}
};

/// Creates a GItem instance in the heap.
GItem& Item(GString const& css_and_text = "");



class GLabel : public GItem {
public:
  GLabel(GString const& css_and_title = "");
  GLabel& clone(GClone const&) override;

  using GItem::operator=;

  using GType = GadgetType_<GLabel,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GLabel instance in the heap.
GLabel& Label(GString const& css_and_title = "");



class GButton : public GItem {
public:
  GButton(GString const& css_and_title = "");
  GButton& clone(GClone const&) override;

  using GItem::operator=;

  using GType = GadgetType_<GButton,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GButton instance in the heap.
GButton& Button(GString const& css_and_title = "");



class GLinkItem : public GItem {
public:
  GLinkItem(GString const& css_and_title = "");
  GLinkItem& clone(GClone const&) override;

  using GItem::operator=;

  using GType = GadgetType_<GLinkItem,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GLinkItem instance in the heap.
GLinkItem& LinkItem(GString const& css_and_title = "");



class GChoiceItem : public GItem {
public:
  GChoiceItem(GString const& css_and_title = "");
  GChoiceItem& clone(GClone const&) override;

  using GItem::operator=;

  using GType = GadgetType_<GChoiceItem,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GChoiceItem instance in the heap.
GChoiceItem& ChoiceItem(GString const& css_and_title = "");



class GCheckItem : public GItem {
public:
  GCheckItem(GString const& css_and_title = "");
  GCheckItem& clone(GClone const&) override;

  using GItem::operator=;

  using GType = GadgetType_<GCheckItem,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GCheckItem instance in the heap.
GCheckItem& CheckItem(GString const& css_and_title = "");



class GRadioItem : public GItem {
public:
  GRadioItem(GString const& css_and_title = "");
  GRadioItem& clone(GClone const&) override;

  using GItem::operator=;

  using GType = GadgetType_<GRadioItem,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GRadioItem instance in the heap.
GRadioItem& RadioItem(GString const& css_and_title = "");



class GTitle : public GItem {
public:
  GTitle(GString const& css_and_title = "");
  GTitle& clone(GClone const&) override;

  using GItem::operator=;
  Place place() const override {return Place::Top;}

  using GType = GadgetType_<GTitle,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GRadioItem instance in the heap.
GTitle& Title(GString const& css_and_title = "");



class GSeparator : public GItem {
public:
  GSeparator(GString const& css = "");
  GSeparator& clone(GClone const&) override;

  using GType = GadgetType_<GSeparator,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GSeparator instance in the heap.
GSeparator& Separator(GString const& css = "");



/// Transparent flexible item for filling up space; see GItem.
class GFiller : public GItem {
public:
  GFiller(GString const& css = "");
  GFiller& clone(GClone const&) override;

  using GType = GadgetType_<GFiller,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a GFiller instance in the heap.
GFiller& Filler(GString const& css = "");

}
#endif
