//
//  Table gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Table_hpp
#define Guit_Table_hpp
#include <gbox.hpp>
namespace guit {

/// Table Box.
/// This gadget roughly works as a HTML table.  A table typically contains a GTHead child
/// (acting as a header row) followed by several GTRow children (acting as rows). A row contains
/// cells (see below), that are vertically aligned.
///
/// A table can also contain other gadgets: children deriving from GBox are considered as rows,
/// other gadgets (e.g. a GLabel or a GSeparator) occupy the full width of the table.
///
/// Cells (the gadgets added to the rows) can be of any Gadget class. A cell can also be a GString
/// or GText, GInt, GFloat, GBool prop. A GString is converted into a GLabel, a GText/GInt/GFloat
/// into a GTextField, and a GBool into a GCheckItem. The prop is bound with the generated gadget
/// (e.g. adding a GInt creates a GTextField that is bound to this GInt).
///
/// A table always has a column direction and its children should have a row direction (see GFlow).
///
class GTable : public GBox {
protected:
  GTable(GadgetType&, GString const& css_and_title);

public:
  ~GTable();
  GTable(GString const& css_and_title = "");
  GTable& clone(GClone const&) override;

  GTable* toTable() override {return this;}

  /// Returns the number of columns.
  /// The returned value is the maximal number or columns (including columns
  /// in THead headers if any).
  /// if _minColumnCount_ is provided it will contain the minimal number or columns.
  int columnCount(int* minColumnCount = 0) const;
  
  /// Returns the number of rows.
  /// This count includes THead headers if any.
  int rowCount() const {return childCount();}
  
  /// Returns the corresponding row.
  /// Returns nullptr if this row does not exist.
  /// Rows start at 0 and THead headers are counted if any.
  Gadget* childAt(int row) const override {return GBox::childAt(row);}
  
  /// Returns the corresponding cell.
  /// Returns nullptr if this cell does not exist.
  /// Rows start at 0 and THead headers are counted if any.
  virtual Gadget* childAt(int row, int column);

  using GBox::perform;

  /// Calls a function on the cells of the table.
  /// _fun_(cell) is called for each cell in the table, except those in THead children.
  void perform(std::function<void(Gadget& cell)> fun) override;
  
  /// Calls a function on the cells of the table.
  /// _fun_(cell, row, column) is called for each cell in the table, except those in THead children.
  /// Rows and columns start at 0. THead headers are counted (ie. if there is a header before
  /// the TRow(s), the first TRow will be at row 1).
  virtual void perform(std::function<void(Gadget& cell, int row, int column)> fun);

  Gadget& createDefaultItem(GString const& arg) override;

  using GType = GadgetType_<GTable,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  using Views = std::vector<GTable*>;
  std::shared_ptr<Views> views_;    // all table instances
  void pack(GRender* parent_render) override;
};


/// Creates a Table Gadget.
GTable& Table(GString const& css_and_title = "");


// - - -  - - -  - - -  - - -  - - -  - - -  - - -  - - -  - - -  - - -  - - -

/// Table Row.
class GTRow : public GBox {
protected:
  GTRow(GadgetType&, GString const& css);

public:
  GTRow(GString const& css = "");

  GTRow& clone(GClone const&) override;
  GTRow* toTRow() override {return this;}

  using GType = GadgetType_<GTRow,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}

  Gadget& createDefaultItem(GString const& arg) override;

protected:
  GDim computeInsideSize(GRender&) override;
  void readValues(std::istream&, GIOBuffer&) override;
  void writeValues(std::ostream&, GIOBuffer&) const override;
};


/// Creates a Table Row.
GTRow& TRow(GString const& css = "");


// - - -  - - -  - - -  - - -  - - -  - - -  - - -  - - -  - - -  - - -  - - -

/// Table Header.
class GTHead : public GTRow {
protected:
  GTHead(GadgetType&, GString const& css);

public:
  GTHead(GString const& css = "");

  GTHead& clone(GClone const&) override;
  GTHead* toTHead() override {return this;}

  using GType = GadgetType_<GTHead,GTRow>;
  static GType& Type();
  GType& type() const override {return Type();}
};


/// Creates a Table Header.
GTHead& THead(GString const& css = "");


}
#endif
