//
//  Text Gadgets: textfield and textarea
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_TextItem_hpp
#define Guit_TextItem_hpp
#include <gitem.hpp>
#include <gtext.hpp>
#include <gfont.hpp>
namespace guit {

/// Base class of gadgets that can edit text.
/// The text is encoded in UTF8. Use Label if you don't need editable text
/// and TextArea if you need multiple lines.
///
/// The text and the associated GText prop can be changed as follows (see GItem):
/// @code
///    auto& tf1 = TextField();
///    tf1 = "Do It";        // changes text of tf1
///    tf1 << "Do It";       // same effect
///    std::cout << tf1;     // prints "Do It"
///
///    auto& txt = Text("Do It");
///    auto& tf2 = TextField();
///    tf2 << txt;           // changes the GText prop of tf2
///    std::cout << tf2;     // prints "Do It"
///    txt = "Whatever";     // changes text of txt and tf2
///    std::cout << tf2;     // prints "Whatever"
/// @endcode
///
/// Props can be shared with other gadgets and bound together.
/// GExpr allows creating textual expressions whose value results from a combination of Types.
/// @see GItem, Gadget, GExpr.
///
class GTextItem : public GItem {
protected:
  GTextItem(GadgetType&, GString const& css_and_value);

public:
  ~GTextItem();
  GTextItem(GString const& css_and_text = "");
  GTextItem& clone(GClone const&) override;

  GTextItem* toTextItem() override {return this;}
  using GItem::add;

  /// Changes the value.
  ///@{
  GTextItem& operator=(GString const& str) override;
  GTextItem& operator=(float) override;
  void setValue(GString const& str) override;
  void add(GString const&) override;
  /// @}

  /// Returns the value as a string.
  /// @{
  GString text() const override;
  //GString const& value() const;
  GString stringValue() const override;
  /// @}

  /// Returns the value as a number.
  /// @{
  virtual int intValue() const;
  virtual float floatValue() const;
  /// @}

  /// Returns the associated GText property.
  /// To change the GText, use operator << :
  /// @code
  ///   auto& txt = Text();
  ///   auto& edit = TextField() << txt;
  /// @endcode
  /// @{
  GText& operator()();
  GProp* valueProp() override {return value_;}
  /// @}

  /// Retreives/changes text attributes.
  /// @{
  void setAttributes(GTextAttributes*);
  GTextAttributes* getAttributes() {return attrlist_;}
  /// @}

  /// Retrieves the position of the caret.
  /// Row and column start at 0 (row is 0 if there is only one row, e.g. in a TextField)
  /// @{
  int  caretColumn();
  void caretRowColumn(int& row, int& column);
  /// @}

  /// Changes the position of the caret.
  /// Line and column start at 0. The row argument is ignored in a TextField.
  /// @{
  void setCaretColumn(int column);
  void setCaretRowColumn(int row, int column);
  /// @}

  /// Returns or changes the index of the caret (position in bytes).
  /// - each UFT8 character can occupy 1 to 4 bytes.
  /// - the argument of setCaretIndex must correspond to the starting byte of a UTF8 character.
  /// @{
  ssize_t caretIndex() const {return ctrl_.caret;}
  void setCaretIndex(ssize_t index);
  /// @}

  /// converts character position to byte position and vice-versa.
  /// - each UTF8 character  can occupy 1 to 4 bytes.
  /// - returns position of last byte/character if first argument is < 0.
  /// @{
  ssize_t posToIndex(ssize_t charpos) const;
  ssize_t indexToPos(ssize_t index) const;
  /// @}

  void setCaret(GPoint const& wpos);
  ssize_t findCaretIndex(GPoint const& wpos);
  void findCaretRowColumn(ssize_t index, int& line, int& column);
  void moveCaret(GPoint const& wpos);
  void moveCaretRight(ssize_t count, bool scroll_one);
  void moveCaretLeft();
  void moveCaretUp();
  void moveCaretDown();
  void insertText(GString const& text);
  void deleteChar(bool before_caret);
  void deleteEnd();
  bool deleteSelection();
  void copyOrCut(bool cut);
  void paste();

  bool hasValue() const override {return true;}

  using GType = GadgetType_<GTextItem,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  struct Ctrl {
    ssize_t offset, caret, selbegin, selend;
  } ctrl_{};
  gptr<GText> value_;
  GFontValue font_;
  const class GMarginValue* margin_{};
  GTextAttributes* attrlist_{};
  void addProp(GProp&, bool add_to_proplist) override;
  void pack(GRender* parent_render) override;
  ssize_t findCaretIndexInRow(ssize_t begin, ssize_t end, float caretX);
  bool keyCB(class GKeyEvent&);
  void selectAllTextCB();
  /// @}
};

/// Creates editable text; see GTextItem.
GTextItem& TextItem(GString const& css_and_text = "");


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class GTextField : public GTextItem {
public:
  GTextField(GString const& css_and_text = "");
  GTextField& clone(GClone const&) override;

  using GTextItem::operator=;

  using GType = GadgetType_<GTextField,GTextItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates an editable text field (single line); see GTextItem.
GTextField& TextField(GString const& css_and_text = "");


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class GTextArea : public GTextItem {
public:
  GTextArea(GString const& css_and_text = "");
  GTextArea& clone(GClone const&) override;

  using GTextItem::operator=;

  using GType = GadgetType_<GTextArea,GTextItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates an editable text area (multilines); see GTextItem.
GTextArea& TextArea(GString const& css_and_text = "");

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Text Attribute.
class GTextAttr {
public:
  GTextAttr(size_t pos_in_text, size_t length);

  size_t pos_{}, len_{};
  GPropList props_;
  gptr<GObject> data_;
};


/// GTextItem attributes (see GTextItem::setAttributes()).
class GTextAttributes {
public:
  GTextAttributes() {list_.clear();}
  void add(GTextAttr&);
  void clear();
  GTextAttr* find(size_t pos) const;
  GTextAttr* find(GObject* data) const;
  void adjust(size_t pos, ssize_t delta);

  std::vector<GTextAttr*> list_;
};

}
#endif
