//
// Animations
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Anim_hpp
#define Guit_Anim_hpp
#include <gtimer.hpp>
namespace guit {

/// Animation Step
class GAnimStep {
public:
  GAnimStep() = default;
  GAnimStep& operator<<(GString const& str) ;
  GAnimStep& operator<<(GProp& p) {list_.push_back(p); return *this;}
  GAnimStep& operator<<(GProp* p) {list_.push_back(p); return *this;}
  bool empty() const {return list_.empty();}
  std::vector<gptr<GProp>> list_;
};

inline GAnimStep& AnimStep() {return *new GAnimStep();}


/// Animation.
/// Blends prop values when applicable, change then in sequence otherwise.
class GAnim : public GProp {
public:
  enum {Infinite = -1, OneShot = -2};

  ~GAnim();
  GAnim(GTime duration = 0, int repetitions = Infinite);

  GAnim(GString const& anim_spec);
  /// anim_spec is in the form:  "duration  repetitions [steps]*"
  /// examples (lasts 5 sec, repeated 3 times):
  /// @code
  ///   "5000 3 {color:red; background:white", color:white; background:red}"
  /// };
  /// anim << step_list;
  /// @endcode

  GAnim* clone(bool) const override {return nullptr;}  ///< can't be cloned.
  GAnim* toAnim() override {return this;}

  bool   isRunning() const;
  int    repetitions() const {return repeat_;}
  GAnim& setRepetitions(int count);
  GTime  duration() const {return duration_;}

  /// starts / stops.
  GAnim& start(bool yes = true);
  void stop() {start(false);}

  /// restarts from the current step.
  void restart();

  /// adds a step with props in textual format.
  /// the argument is a ; separated list of props, examples:
  /// @code
  ///    anim << "color: red";
  ///    anim << "pos: 200 200; color: red";
  /// @endcode
  GAnim& operator<<(GString const& step_props);

  /// adds a step list with props in textual format.
  /// example:
  /// @code
  /// GStrings step_list{
  ///   "pos: 200 200; color: red",
  ///   "pos: 300 300; color: orange",
  ///   "pos: 200 200; color: none"
  /// };
  /// anim << step_list;
  /// @endcode
  GAnim& operator<<(GStrings const& step_list);

  /// adds a step (only one GProp).
  /// example:
  /// @code
  ///    anim << GColor::red;
  /// @endcode
  /// @{
  GAnim& operator<<(GProp& prop);
  GAnim& operator<<(GProp* prop);
  /// @}

  /// adds a step (multiple GProp(s)).
  /// example:
  /// @code
  /// anim
  /// << (GAnimStep() << Pos(200,200) << Color("red"))
  /// << (GAnimStep() << Pos(300,200) << Color("orange"))
  /// << (GAnimStep() << Pos(200,200) << Color("none"));
  /// @endcode
  GAnim& operator<<(GAnimStep& step);

  /// adds a callback function which is fired at each step.
  GAnim& operator<<(GFunction const& fun);

  GTimer* timer() {return timer_;}
  GString stringValue() const override;

  using GType = GPropType_<GAnim,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

  bool isEquivalent(GProp const& other) const override {return false;}
  void apply(Gadget*, GRender&, Specif) override;
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;

protected:
  friend class GVarProp;
  virtual void init();
  void timeoutCB();
  void fireCallbacks(GEvent&);
  GTime duration_{}, elapsed_{}, timeout_{};
  int  repeat_{1};
  bool startdelayed_{};
  gptr<GTimer> timer_;
  GPropList calls_;
  GAnimStep props_;
  std::vector<GAnimStep*> steps_;
  void updateProps(GAnimStep& from, GAnimStep& to, float alpha);
};

/// Returns a GAnim created in the heap.
inline GAnim& Anim(GTime duration = 0, int count = GAnim::Infinite) {
  return *new GAnim(duration, count);
}

}
#endif
