//
// Color and Background Props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Color_hpp
#define Guit_Color_hpp
#include <gvarprop.hpp>
namespace guit {

/// RGBA Color.
/// @see GColor, which is a gadget property.
class GRgba {
public:
  GRgba() : red{}, green{}, blue{}, alpha{1.f} {}
  GRgba(float red, float green, float blue, float alpha = 1.f);
  GRgba(GString const& colorname);

  void set(float red, float green, float blue, float alpha = 1.f);
  bool set(GString const& colorname);

  /// returns alpha * this_color + (1-alpha) * color2
  static GRgba blend(const GRgba& color1, const GRgba& color2, float alpha);

  GString stringValue() const;

  bool operator==(const GRgba&) const;
  bool operator!=(const GRgba&) const;

  friend std::ostream& operator<<(std::ostream& out, const GRgba&);
  friend std::istream& operator>>(std::istream& in, GRgba&);

  float red, green, blue, alpha;
};

inline bool gconvert(const GRgba& from, GRgba& to) {to = from; return true;}
bool gconvert(GString const& from, GRgba& to);
bool gconvert(GRgba const& from, GString& to);

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Foreground color of a gadget.
/// @note Text can also be attributed using a pseudo-html syntax, see Gadget::setRichText().
class GColor : public GVarProp_<GRgba> {
public:
  using Super = GVarProp_<GRgba>;

  /// Predefined colors.
  static GColor none, white, black, gray, lightgray, dimgray,
  red, green, blue, lightblue, guitblue, cyan, magenta, yellow, orange;

  GColor();

  /// Creates a color from a RGB or RGBA values.
  /// values must be between  0.f and 1.f.
  GColor(float red, float green, float blue, float alpha = 1.f);

  /// Create a color from a GRgba struct.
  /// values must be between  0.f and 1.f.
  GColor(GRgba const& color);

  /// Create a color from another color.
  GColor(GColor const& color);

  /// Create a color from another color with a different alpha.
  GColor(GColor const& color, float alpha);

  /// Creates a color from a color name.
  /// _name_ can either be:
  /// - a CSS color name such as "AliceBlue" (case unsensitive)
  /// - an hexadecimal color as: "#RRGGBB" or "#RRGGBBAA" (case unsensitive)
  /// - a RGB or RGBA color as: "r,g,b" or "r,g,b,a" where r,g,b,a are floats
  ///   between 0 and 1.
  GColor(GString const& value);

  GColor* clone(bool copy_value) const override;

  GColor& operator=(GRgba const& value);
  GColor& operator=(GColor const& color);
  GColor& operator=(GString const& value);

  using GVarProp_<GRgba>::set;
  void set(float red, float green, float blue, float alpha = 1.0f);

  void blend(const GColor& a, const GColor& b, float mix);
  bool blend(const GProp& a, const GProp& b, float mix) override;

  void setRed(float);
  void setGreen(float);
  void setBlue(float);
  void setAlpha(float);

  float redValue()   const {return value_.red;}
  float blueValue()  const {return value_.blue;}
  float greenValue() const {return value_.green;}
  float alphavalue() const {return value_.alpha;}

  /// Binds this prop to a textual active expression.
  GColor& operator<<=(class GTextExpr&);

  /// Future value in an Active Expression.
  /// An Active Expression (GExpr) can refer ~prop.
  auto operator~() {return GPropRef_<GColor>(*this);}

  /// Field of a GColor prop.
  using Field = GNumPropField<GColor,float>;

  /// Returns a Field for active expressions or binding.
  /// This field can be bound to a prop or another prop field using <<=.
  /// A textual Active Expression (GExpr) can refer this field.
  /// @{
  Field redField()   {return Field(*this, value_.red);}
  Field greenField() {return Field(*this, value_.green);}
  Field blueField()  {return Field(*this, value_.blue);}
  Field alphaField() {return Field(*this, value_.alpha);}
  /// @}

  using GType = GPropType_<GColor,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal.
  /// @{
  bool isValid(GRgba const&) const override;
  bool doChange(GRgba const&) override;
  void apply(Gadget*, GRender&, Specif) override;
  void updateGadget(Gadget&) override;
  /// @}
};

/// Return a GColor created in the heap.
/// @{
inline GColor& Color() {return *new GColor();}
inline GColor& Color(float r, float g, float b, float a = 1.f) {return *new GColor(r,g,b,a);}
inline GColor& Color(GString const& value) {return *new GColor(value);}
/// @}


using BackgroundAttr = GRgba;

/// Background color of a gadget.
class GBackground : public GColor {
public:
  using Super = GColor;
  
  /// Predefined Background colors.
  static GBackground none, white, black, gray, lightgray, dimgray,
  red, green, blue, lightblue, guitblue, cyan, magenta, yellow, orange;

  GBackground();
  GBackground(float red, float green, float blue, float alpha = 1.f);
  GBackground(GRgba const& color);
  GBackground(GColor const& color);
  GBackground(GColor const& color, float alpha);
  GBackground(GString const& name);

  GBackground* clone(bool copy_value) const override;

  GBackground& operator=(GRgba const& value);
  GBackground& operator=(GColor const& value);
  GBackground& operator=(GString const& value);

  using GType = GPropType_<GBackground,GColor>;
  static GType& Type();
  GType& type() const override {return Type();}

  void apply(Gadget*, GRender&, Specif) override;
};

/// Return a GBackground created in the heap.
/// @{
inline GBackground& Background() {return *new GBackground();}
inline GBackground& Background(float r, float g, float b, float a = 1.f) {return *new GBackground(r,g,b,a);}
inline GBackground& Background(GString const& value) {return *new GBackground(value);}
/// @}

}
#endif
