//
// List of Props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitPropList_hpp
#define GuitPropList_hpp
#include <forward_list>
#include <gprop.hpp>
#include <gptr.hpp>
namespace guit {

/// List of props.
/// The list owns its arguments, they won't be deleted while the list exists.
class GPropList {
public:
  using iterator = std::forward_list<gptr<GProp>>::iterator;

  GPropList() = default;
  GPropList(std::initializer_list<gptr<GProp>> prop_list);

  auto& list() {return list_;}
  const auto& list() const {return list_;}
  size_t size() const;
  bool empty() const {return list_.empty();}

  iterator begin() {return list_.begin();}
  iterator before_begin() {return list_.before_begin();}
  iterator end() {return list_.end();}

  auto begin() const {return list_.begin();}
  auto before_begin() const {return list_.before_begin();}
  auto end() const {return list_.end();}

  GProp* findProp(GPropType&) const;
  template <class T> T* findProp() const {return dynamic_cast<T*>(findProp(T::Class()));}
  GProp* findProp(std::function<bool(GProp&)>& predicate) const;

  /// add or replace prop.
  /// searches prop with same role in the list, replaces it if found, adds it to the end otherwise
  void addReplace(GProp&, Gadget*);

  /// add prop to the beginning or the ned of the list (without checking if already in the list).
  /// @{
  void addBegin(GProp&);
  void addBegin(GPropList const&);
  void addEnd(GProp&);
  void addEnd(GPropList const&);
  /// @}

  bool remove(Gadget*, std::function<bool(GProp&)> pred);
  bool remove(Gadget*, GProp&);
  bool remove(Gadget*, GPropType&);
  void removeAll(Gadget*);

  void fire(GEvent&);
  void fire(GObject* aux);

  /// returns first notifier refering to obj.
  class GNotifier* findNotifier(void* obj);

  /// removes all notifiers refering obj.
  void removeNotifiers(void* obj);

private:
  friend class Gadget;
  friend class GVarProp;
  std::forward_list<gptr<GProp>> list_;
};

}
#endif
