//
// internals
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_PropComp_hpp
#define Guit_PropComp_hpp
#include <sstream>
#include <gtypes.hpp>
namespace guit {

class GFieldSpec {
public:
  using Value = void;
  bool isoptional_{};

  virtual ~GFieldSpec() = default;
  virtual bool toValue(GString const& from, Value* to) = 0;
  virtual bool toString(const Value* from, GString& to) = 0;
};


class GPropSpec {
protected:
  std::vector<GFieldSpec*> fields_;
public:
  using Value = void;

  GPropSpec(std::initializer_list<GFieldSpec*> fields);
  int toValue(GString const& str, Value*);
  GString toString(const Value*);
};


// - - - specializations

template <class T, class Var>
class GNumFieldSpec : public GFieldSpec {
  Var T::* var_{};
public:
  GNumFieldSpec(Var T::* var, bool isoptional = false)
  : var_(var) {isoptional_ = isoptional;}

  bool toValue(GString const& from, Value* to) override {
    return gconvert(from, ((T*)to)->*var_);
  }

  bool toString(const Value* from, GString& to) override {
    return gconvert(((const T*)from)->*var_, to);
  }
};

// - - - - -

class GFieldSpecValue {
public:
  GString name_;
  int type_{};
  bool isalias_{};
  float value_{};
  uint64_t nameid_{};
  bool islabel_{};
};

class GFieldSpecValues {
public:
  std::vector<GFieldSpecValue> list_;

  GFieldSpecValues(std::initializer_list<GFieldSpecValue> values);
  GFieldSpecValue const* find1(GString const& str) const;
  GFieldSpecValue const* find2(GString const& str, float&) const;
  bool find1(int type, GString& to) const;
  bool find2(int type, float val, GString& to) const;
};


template <class T, class Var>
class GListFieldSpec : public GFieldSpec {
  Var T::* var_{};
  GFieldSpecValues const& values_;

public:
  // NOTE: values are not copied!
  GListFieldSpec(Var T::* var, GFieldSpecValues const& values)
  : var_(var), values_(values) {}

  bool toValue(GString const& str, Value* to) override {
    if (auto* it = values_.find1(str)) {
      ((T*)to)->*var_ = (Var) it->type_;
      return true;
    }
    else return false;
  }

  bool toString(const Value* from, GString& to) override {
    return values_.find1(((const T*)from)->*var_, to);
  }
};


template <class T1, class T2, class Var1, class Var2>
class GListFieldSpec2 : public GFieldSpec  {
  Var1 T1::* var1_{};
  Var2 T2::* var2_{};
  GFieldSpecValues const& values_;

public:
  // NOTE: values are not coped!
  GListFieldSpec2(Var1 T1::* var1, Var2 T2::* var2, GFieldSpecValues const& values)
  : var1_(var1), var2_(var2), values_(values) {}

  bool toValue(GString const& str, Value* to) override {
    float val{};
    if (auto* it = values_.find2(str, val)) {
      ((T1*)to)->*var1_ = Var1(it->type_);
      ((T2*)to)->*var2_ = Var2(val);
      return true;
    }
    else return false;
  }

  bool toString(const Value* from, GString& to) override {
    return values_.find2(((const T1*)from)->*var1_, ((T2*)from)->*var2_, to);
  }
};

}
#endif
